# Copyright (c) 2020 Stefan Neumann and Pauli Miettinen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import copy
from scipy.stats import rankdata

try:
    import centermodule
except ImportError as e:
    raise ImportError("You need to compile the centermodule before running this script; see README for more information") from e


import HelperFunctions

# Possible covering options are 'FindOne' and 'FindAllAssoStyleTopK'.
# Use 'FindOne' to get the biclustering-style covering as described in Section 4.1
# of the paper.
# Use 'FindAllAssoStyleTopK' to get the BMF-/asso-style covering as described in
# Section 4.2 of the paper.
def coverLeftSide(filepath,
				  rightClusters,
				  coveringType='FindAllAssoStyleTopK',
				  minOverlap=0.0, # see main.py for an explanation of minOverlap
				  fileIsSparse=False,
				  skipHeader=False,
				  k=None
	):
	if coveringType not in ['FindAllAssoStyle', 'FindAllAssoStyleTopK', 'FindOne']:
		print(f'Unknown coveringType {coveringType} passed. Not covering the left side.')
		return []

	if k == None:
		k = len(rightClusters)

	leftClusters = []
	numClusters = len(rightClusters)
	for i in range(len(rightClusters)):
		leftClusters.append( [] )

	row = -1
	gains = [0] * len(rightClusters)

	rightClustersAsList = []
	for cluster in rightClusters:
		rightClustersAsList.append( sorted(cluster) )
	with open(filepath) as fp:
		if fileIsSparse and skipHeader:
			fp.readline()

		while True:
			row += 1
			rowIdx = row
			if fileIsSparse:
				p,rowIdx = HelperFunctions.getNextPointSparseInput(fp)
			else:
				p,curCols = HelperFunctions.getNextPointDenseInput(fp)

			if len(p) == 0:
				break

			p = set(p)
			if coveringType in ['FindAllAssoStyle', 'FindAllAssoStyleTopK']:
				addedClusterToCover = True
				
				toCover = copy.deepcopy(p)
				overcovered = set()
				pOrOvercovered = copy.deepcopy(p)
				while addedClusterToCover:
					addedClusterToCover = False

					bestClusterIndex, bestScore = centermodule.clusterWithBestGain(sorted(toCover), sorted(pOrOvercovered), rightClustersAsList)
					'''
					# IF YOU DO NOT WANT TO COMPILE CENTERMODULE, YOU CAN USE
					# THE FOLLOWING PYTHON CODE, BUT THIS WILL BE MUCH SLOWER

					bestClusterIndex = -1
					bestScore = 0

					for i in range(len(rightClusters)):
						cluster = set(rightClusters[i])
						if len(cluster) == 0:
							continue

						newlyCovered = cluster & toCover
						newlyOvercovered = cluster - (p | overcovered)

						score = len(newlyCovered) - len(newlyOvercovered)
						
						if score > bestScore:
							bestClusterIndex = i
							bestScore = score
					'''
							
					if bestScore > 0:
						addedClusterToCover = True

						leftClusters[bestClusterIndex].append(rowIdx)
						bestCluster = set(rightClusters[bestClusterIndex])
						toCover.difference_update(bestCluster)
						overcovered = overcovered | (bestCluster - p)
						pOrOvercovered = p | overcovered
						gains[bestClusterIndex] += bestScore
			elif coveringType == 'FindOne': # FindOne, this is the option for the biclustering algorithm described in the paper
				highestOverlap = 0
				clusterWithHighestOverlap = -1
				
				for i in range(len(rightClusters)):
					cluster = set(rightClusters[i])
					if len(cluster) == 0:
						continue

					overlappingIndices = p.intersection(cluster)
					overlap = len(overlappingIndices) / len(cluster)

					if coveringType == 'FindOne':
						if overlap > highestOverlap and overlap >= minOverlap:
							highestOverlap = overlap
							clusterWithHighestOverlap = i

				if clusterWithHighestOverlap > -1:
					leftClusters[clusterWithHighestOverlap].append(rowIdx)
			else:
				print('Unknown left-side covering option presented.')
				return

	if coveringType == 'FindAllAssoStyleTopK':
		ranks = rankdata(gains, method='ordinal')
		clustersToKeep = []

		for i in range(len(leftClusters)):
			rank = ranks[i]
			if rank <= len(leftClusters) - k: # ranks are 1-indexed
				leftClusters[i] = []

	return rightClusters, leftClusters

