/*
 * compile with
 * python3 setup.py build && python3 setup.py install
 */
/*
 * Copyright (c) 2020 Stefan Neumann and Pauli Miettinen
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <Python.h>

#include <limits.h>
#include <float.h>

double
_distanceOfPointAndCenter(PyObject* p, PyObject* center, double asymHammingDistanceWeight)
{
	double dist = 0;
	long i = 0;

	long lengthP = PyObject_Length(p);
	long centerLength = PyObject_Length(center);

	if (centerLength == 0) {
		return lengthP;
	}

	long p_j;
	long center_i = PyLong_AsLong( PyList_GetItem(center, i) );

	for (long j=0; j<lengthP; j++) {
		p_j = (long) PyFloat_AsDouble( PyList_GetItem(p, j) );
		while (i < centerLength && center_i < p_j) {
			i += 1;
			if (i < centerLength) {
				center_i = PyLong_AsLong( PyList_GetItem(center, i) );
			}
			dist += asymHammingDistanceWeight; // center_i is center but not in p
		}

		// we finished processing center and all entries left in p must increase the distance
		if (i >= centerLength) {
			dist += lengthP - j; // the entries are in p but not in center
			break;
		}

		if (center_i != p_j) {
			dist += 1; // p_j is in p but not in center
		} else {
			i += 1;
		}
	}

	if (i < centerLength) {
		dist += asymHammingDistanceWeight*(centerLength - i); // the entries are in center but not in p
	}
	
	return dist;
}


double
_closestCenter(PyObject* p, PyObject* centers, long* centerIndex, double asymHammingDistanceWeight)
{
	double minDistance = DBL_MAX;

	long numCenters = PyObject_Length(centers);

	for (long i=0; i<numCenters; i++) {
		PyObject* center = PyList_GetItem(centers, i);
		double dist = _distanceOfPointAndCenter(p, center, asymHammingDistanceWeight);

		if (dist < minDistance) {
			minDistance = dist;
			*centerIndex = i;
		}
	}

	return minDistance;
}

static PyObject*
closestCenter(PyObject* self, PyObject *args)
{
	PyObject *centers, *p;
	double asymHammingDistanceWeight;
	PyArg_ParseTuple(args, "O!O!d", &PyList_Type, &centers, &PyList_Type, &p, &asymHammingDistanceWeight);

	long centerIndex = -1909;
	double distance = _closestCenter(p, centers, &centerIndex, asymHammingDistanceWeight);

	return Py_BuildValue("dl", distance, centerIndex);
}

long
_clusterWithBestGain(PyObject* toCover,
					 PyObject* pOrOvercovered,
					 PyObject* clusters,
					 long* bestClusterIndex)
{
	long bestScore = 0;
	*bestClusterIndex = -1;

	long lengthToCover = PyObject_Length(toCover);
	long lengthPOrOvercovered = PyObject_Length(pOrOvercovered);

	if (lengthToCover == 0) {
		return 0;
	}

	long numClusters = PyObject_Length(clusters);

	for (long i=0; i<numClusters; i++) {
		PyObject* cluster = PyList_GetItem(clusters, i);
		long clusterLength = PyObject_Length(cluster);

		if (clusterLength == 0) {
			continue;
		}

		long score = 0;

		long toCoverIndex = 0;
		long pOrOvercoveredIndex = 0;
		long toCoverElement = PyLong_AsLong( PyList_GetItem(toCover, 0) );
		long pOrOvercoveredElement = PyLong_AsLong( PyList_GetItem(pOrOvercovered, 0) );

		for (long j=0; j<clusterLength; j++) {
			long item = PyLong_AsLong( PyList_GetItem(cluster, j) );
			while (toCoverIndex < lengthToCover && toCoverElement < item) {
				toCoverIndex += 1;
				if (toCoverIndex < lengthToCover) {
					toCoverElement = PyLong_AsLong( PyList_GetItem(toCover, toCoverIndex) );
				}
			}
			while (pOrOvercoveredIndex < lengthPOrOvercovered && pOrOvercoveredElement < item) {
				pOrOvercoveredIndex += 1;
				if (pOrOvercoveredIndex < lengthPOrOvercovered) {
					pOrOvercoveredElement = PyLong_AsLong( PyList_GetItem(pOrOvercovered, pOrOvercoveredIndex) );
				}
			}

			if (item == toCoverElement) {
				score += 1; // we are covering
			} else if (item != pOrOvercoveredElement) {
				score -= 1; // we are overcovering
			}
		}

		if (score > bestScore) {
			*bestClusterIndex = i;
			bestScore = score;
		}
	}

	return bestScore;
}

static PyObject*
clusterWithBestGain(PyObject* self, PyObject *args)
{
	PyObject *toCover, *pOrOvercovered, *clusters;
	PyArg_ParseTuple(args, "O!O!O!", &PyList_Type, &toCover, &PyList_Type, &pOrOvercovered, &PyList_Type, &clusters);

	long bestClusterIndex = -1;
	long bestScore = _clusterWithBestGain(toCover,
					 pOrOvercovered,
					 clusters,
					 &bestClusterIndex);

	return Py_BuildValue("ll", bestClusterIndex, bestScore);
}

static PyMethodDef CenterMethods[] = {
    {"closestCenter",  closestCenter, METH_VARARGS, "FindClosestCenter"},
    {"clusterWithBestGain", clusterWithBestGain, METH_VARARGS, "ClusterWithBestGain"},
    {NULL, NULL, 0, NULL}        /* Sentinel */
};

static struct PyModuleDef centermodule = {
    PyModuleDef_HEAD_INIT,
    "centermodule",   /* name of module */
    NULL, /* module documentation, may be NULL */
    -1,       /* size of per-interpreter state of the module,
                 or -1 if the module keeps state in global variables. */
    CenterMethods
};

PyMODINIT_FUNC
PyInit_centermodule(void)
{
    return PyModule_Create(&centermodule);
}

int
main(int argc, char *argv[])
{
    wchar_t *program = Py_DecodeLocale(argv[0], NULL);
    if (program == NULL) {
        fprintf(stderr, "Fatal error: cannot decode argv[0]\n");
        exit(1);
    }

    /* Add a built-in module, before Py_Initialize */
    PyImport_AppendInittab("centermodule", PyInit_centermodule);

    /* Pass argv[0] to the Python interpreter */
    Py_SetProgramName(program);

    /* Initialize the Python interpreter.  Required. */
    Py_Initialize();

    /* Optionally import the module; alternatively,
       import can be deferred until the embedded script
       imports it. */
    PyImport_ImportModule("centermodule");

    PyMem_RawFree(program);
    return 0;
}

